import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from sublayer import MultiHeadAttention
from config2 import PAD


class PositionwiseFeedForward(nn.Module): 
    def __init__(self, d_in, d_hid, dropout=0.1):
        super().__init__()
        self.w_1 = nn.Conv1d(d_in, d_hid, 1) 
        self.w_2 = nn.Conv1d(d_hid, d_in, 1)
        self.layer_norm = nn.LayerNorm(d_in)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        residual = x
        output = x.transpose(1, 2)
        output = self.w_2(F.relu(self.w_1(output)))
        output = output.transpose(1, 2)
        output = self.dropout(output)
        output = self.layer_norm(output + residual) 
        return output


def get_sinusoid_encoding_table(n_position, d_hid, padding_idx=None): 

    def cal_angle(position, hid_idx): 
        return position / np.power(10000, 2 * (hid_idx // 2) / d_hid)

    def get_posi_angle_vec(position): 
        return [cal_angle(position, hid_j) for hid_j in range(d_hid)]

    sinusoid_table = np.array([get_posi_angle_vec(pos_i) for pos_i in range(n_position)]) 

    sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2])  
    sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2])  

    if padding_idx is not None: 
        sinusoid_table[padding_idx] = 0.

    return torch.FloatTensor(sinusoid_table)


def get_non_pad_mask(seq): 
    assert seq.dim() == 2 
    return seq.ne(PAD).type(torch.float).unsqueeze(-1)


def get_subsequent_mask(seq): 

    sz_b, len_s = seq.size() 
    subsequent_mask = torch.triu(
        torch.ones((len_s, len_s), device=seq.device, dtype=torch.uint8), diagonal=1)
    subsequent_mask = subsequent_mask.unsqueeze(0).expand(sz_b, -1, -1) 

    return subsequent_mask


def get_attn_key_pad_mask(seq_k, seq_q): 

    len_q = seq_q.size(1) 
    padding_mask = seq_k.eq(PAD)
    padding_mask = padding_mask.unsqueeze(1).expand(-1, len_q, -1)
    return padding_mask

class EncoderLayer(nn.Module):
    def __init__(self, d_model, d_inner, n_head, d_k, d_v, dropout=0.1): 
        super(EncoderLayer, self).__init__()
        self.slf_attn = MultiHeadAttention(
            n_head, d_model, d_k, d_v, dropout=dropout)
        self.pos_ffn = PositionwiseFeedForward(d_model, d_inner, dropout=dropout) 

    def forward(self, enc_input, non_pad_mask=None, slf_attn_mask=None): 
        enc_output, enc_slf_attn = self.slf_attn( 
            enc_input, enc_input, enc_input, mask=slf_attn_mask)
        enc_output *= non_pad_mask 

        enc_output = self.pos_ffn(enc_output)
        enc_output *= non_pad_mask 

        return enc_output, enc_slf_attn